# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-26 17:17
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0011_auto_20170326_1955'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='collection',
            name='exclude_properties_filters',
        ),
    ]



from cms.app_base import CMSApp
from cms.apphook_pool import apphook_pool
from django.utils.translation import ugettext_lazy as _
from .cms_menus import CollectionMenu


class CollectionApphook(CMSApp):
    app_name = "pcart_collection"
    name = _("Product collection")
    menus = [CollectionMenu]

    def get_urls(self, page=None, language=None, **kwargs):
        return ["pcart_catalog.urls_collection"]

apphook_pool.register(CollectionApphook)


class ProductApphook(CMSApp):
    app_name = "pcart_product"
    name = _("Product detail")

    def get_urls(self, page=None, language=None, **kwargs):
        return ["pcart_catalog.urls_product"]

apphook_pool.register(ProductApphook)




# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-09 19:30
from __future__ import unicode_literals

from django.db import migrations

SQL_1 = '''
CREATE OR REPLACE FUNCTION process_commit_product() RETURNS TRIGGER AS $pcart_commit_product$
    DECLARE tagset varchar[];
    DECLARE variants_tagset varchar[];
    DECLARE merged_tagset varchar[];
    DECLARE max_status_id uuid;
    BEGIN
        -- Build a set of tags for product without variants.
        IF (NEW.variants_count = 0) THEN
            tagset := (SELECT ARRAY(
                SELECT
                    concat(ptp.tag_prefix::text, ':'::text, lower(json_data.value)::text)
                FROM
                    pcart_catalog_product AS p,
                    jsonb_each_text(p.properties) AS json_data
                INNER JOIN pcart_catalog_producttypeproperty AS ptp ON
                    json_data.key = ptp.title AND
                    json_data.value <> '' AND
                    ptp.tag_prefix <> '' AND
                    ptp.product_type_id = (SELECT t.product_type_id FROM pcart_catalog_product AS t WHERE t.id = NEW.id)
                WHERE p.id = NEW.id));

            -- Update tags for itself.
            IF (NEW.tags <> tagset) THEN
                UPDATE pcart_catalog_product
                SET tags = tagset
                WHERE id = NEW.id;
            END IF;

            -- Update price
            IF (NEW.price <> NEW.min_variant_price OR NEW.price <> NEW.max_variant_price) THEN
                UPDATE pcart_catalog_product
                SET min_variant_price = NEW.price, max_variant_price = NEW.price
                WHERE id = NEW.id;
            END IF;
            RETURN NEW;
        ELSE
            tagset := (SELECT ARRAY(
                SELECT
                    concat(ptp.tag_prefix::text, ':'::text, lower(json_data.value)::text)
                FROM
                    pcart_catalog_product AS p,
                    jsonb_each_text(p.properties) AS json_data
                INNER JOIN pcart_catalog_producttypeproperty AS ptp ON
                    json_data.key = ptp.title AND
                    json_data.value <> '' AND
                    ptp.tag_prefix <> '' AND
                    ptp.product_type_id = (SELECT t.product_type_id FROM pcart_catalog_product AS t WHERE t.id = NEW.id)
                WHERE p.id = NEW.id));

            variants_tagset := (
                SELECT ARRAY(
                    SELECT DISTINCT unnest(v.tags) AS tag
                    FROM pcart_catalog_productvariant AS v
                    WHERE v.product_id = NEW.id));

            -- Merge two sets of tags
            merged_tagset := (SELECT ARRAY(SELECT DISTINCT unnest(tagset || variants_tagset)));

            IF (NEW.tags <> merged_tagset) THEN
                UPDATE pcart_catalog_product
                SET tags = merged_tagset
                WHERE id = NEW.id;
            END IF;

            -- Set status as maximum for all its variants.
            max_status_id := (SELECT id FROM pcart_catalog_productstatus WHERE weight = (
	            SELECT max(status.weight)
	            FROM pcart_catalog_productvariant AS v
	            INNER JOIN pcart_catalog_productstatus AS status ON v.status_id = status.id
	            WHERE v.product_id = NEW.id));
	        IF (NEW.status_id <> max_status_id) THEN
                UPDATE pcart_catalog_product
                SET status_id = max_status_id
                WHERE id = NEW.id;
            END IF;
            RETURN NEW;
        END IF;
        RETURN NULL; -- result is ignored since this is an AFTER trigger
    END;
$pcart_commit_product$ LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS pcart_commit_product ON pcart_catalog_product;
CREATE TRIGGER pcart_commit_product
AFTER INSERT OR UPDATE ON pcart_catalog_product
    FOR EACH ROW EXECUTE PROCEDURE process_commit_product();
'''

SQL_2 = '''
CREATE OR REPLACE FUNCTION process_agg_variants() RETURNS TRIGGER AS $pcart_agg_variants$
    DECLARE tagset varchar[];
    DECLARE product_tagset varchar[];
    DECLARE variants_tagset varchar[];
    BEGIN
        -- We need the similar behaviour both for INSERT and UPDATE.
        IF (TG_OP = 'INSERT' OR TG_OP = 'UPDATE') THEN
            -- Build an array of tags of all variants.
            tagset := (SELECT ARRAY(
                SELECT
                    concat(ptp.tag_prefix::text, ':'::text, lower(json_data.value)::text)
                FROM
                    pcart_catalog_productvariant AS v,
                    jsonb_each_text(v.properties) AS json_data
                INNER JOIN pcart_catalog_producttypeproperty AS ptp ON
                    json_data.key = ptp.title AND
                    json_data.value <> '' AND
                    ptp.tag_prefix <> '' AND
                    ptp.product_type_id = (
                        SELECT t.product_type_id FROM pcart_catalog_product AS t WHERE t.id = NEW.product_id)
                WHERE v.id = NEW.id));

            -- Build an array of tags for parent product (for common properties).
            product_tagset := (SELECT ARRAY(
                SELECT
                    concat(ptp.tag_prefix::text, ':'::text, lower(json_data.value)::text)
                FROM
                    pcart_catalog_product AS p,
                    jsonb_each_text(p.properties) AS json_data
                INNER JOIN pcart_catalog_producttypeproperty AS ptp ON
                    json_data.key = ptp.title AND
                    json_data.value <> '' AND
                    ptp.tag_prefix <> '' AND
                    ptp.product_type_id = (
                        SELECT t.product_type_id FROM pcart_catalog_product AS t WHERE t.id = NEW.product_id)
                WHERE p.id = NEW.product_id));

            -- Setup tags for variant. The next operator is necessary for prevent a recursion.
            IF (NEW.tags <> tagset) THEN
                UPDATE pcart_catalog_productvariant SET tags = tagset WHERE id = NEW.id;
            END IF;

            -- Update tags for parent product.
            variants_tagset := (
                SELECT ARRAY(
                    SELECT DISTINCT unnest(v.tags) AS tag
                    FROM pcart_catalog_productvariant AS v
                    WHERE v.product_id = NEW.product_id));

            -- Merge two sets of tags
            UPDATE pcart_catalog_product
            SET tags = (SELECT ARRAY(SELECT DISTINCT unnest(product_tagset || variants_tagset)))
            WHERE id = NEW.product_id;

            -- Update variants count, min/max variant prices.
            UPDATE pcart_catalog_product SET
                (variants_count, min_variant_price, max_variant_price) =
			(SELECT COUNT(*), MIN(v.price), MAX(v.price)
			    FROM pcart_catalog_productvariant AS v WHERE v.product_id = NEW.product_id)
		    WHERE id = NEW.product_id;
            RETURN NEW;
        ELSIF (TG_OP = 'DELETE') THEN
            -- Build an array of tags of all variants.
            tagset := (SELECT ARRAY(
                SELECT
                    concat(ptp.tag_prefix::text, ':'::text, lower(json_data.value)::text)
                FROM
                    pcart_catalog_productvariant AS v,
                    jsonb_each_text(v.properties) AS json_data
                INNER JOIN pcart_catalog_producttypeproperty AS ptp ON
                    json_data.key = ptp.title AND
                    json_data.value <> '' AND
                    ptp.tag_prefix <> '' AND
                    ptp.product_type_id = (
                        SELECT t.product_type_id FROM pcart_catalog_product AS t WHERE t.id = OLD.product_id)
                WHERE v.id = OLD.id));

            -- Build an array of tags for parent product (for common properties).
            product_tagset := (SELECT ARRAY(
                SELECT
                    concat(ptp.tag_prefix::text, ':'::text, lower(json_data.value)::text)
                FROM
                    pcart_catalog_product AS p,
                    jsonb_each_text(p.properties) AS json_data
                INNER JOIN pcart_catalog_producttypeproperty AS ptp ON
                    json_data.key = ptp.title AND
                    json_data.value <> '' AND
                    ptp.tag_prefix <> '' AND
                    ptp.product_type_id = (
                        SELECT t.product_type_id FROM pcart_catalog_product AS t WHERE t.id = OLD.product_id)
                WHERE p.id = OLD.product_id));

            -- Update tags for parent product.
            variants_tagset := (
                SELECT ARRAY(
                    SELECT DISTINCT unnest(v.tags) AS tag
                    FROM pcart_catalog_productvariant AS v
                    WHERE v.product_id = OLD.product_id));

            -- Merge two sets of tags
            UPDATE pcart_catalog_product
            SET tags = (SELECT ARRAY(SELECT DISTINCT unnest(product_tagset || variants_tagset)))
            WHERE id = OLD.product_id;

            -- Update variants count, min/max variant prices.
            UPDATE pcart_catalog_product SET
                (variants_count, min_variant_price, max_variant_price) =
			(SELECT COUNT(*), COALESCE(MIN(v.price), 0), COALESCE(MAX(v.price), 0)
			    FROM pcart_catalog_productvariant AS v WHERE v.product_id = OLD.product_id)
		    WHERE id = OLD.product_id;
		    RETURN OLD;
        END IF;
        RETURN NULL; -- result is ignored since this is an AFTER trigger
    END;
$pcart_agg_variants$ LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS pcart_agg_variants ON pcart_catalog_productvariant;
CREATE TRIGGER pcart_agg_variants
AFTER INSERT OR UPDATE OR DELETE ON pcart_catalog_productvariant
    FOR EACH ROW EXECUTE PROCEDURE process_agg_variants();
'''


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0001_initial'),
    ]

    operations = [
        migrations.RunSQL(SQL_1),
        migrations.RunSQL(SQL_2),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-12 17:31
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
        ('pcart_catalog', '0008_productimage_downloaded'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionPluginModel',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='pcart_catalog_collectionpluginmodel', serialize=False, to='cms.CMSPlugin')),
                ('filter_string', models.CharField(blank=True, default='', max_length=255, verbose_name='Filter string')),
                ('sorting', models.CharField(blank=True, default='cheap', max_length=100, verbose_name='Sorting')),
                ('template_name', models.CharField(default='plugin', max_length=100, verbose_name='Template name')),
                ('collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='pcart_catalog.Collection', verbose_name='Collection')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
    ]



from django.db import models
from django.contrib.postgres.fields import JSONField, ArrayField
from django.utils.translation import ugettext_lazy as _
from django.contrib.sites.models import Site
from django.urls import reverse, NoReverseMatch
from django.db.models.signals import pre_save, post_save, post_delete
from django.dispatch import receiver
from django.conf import settings
from django.utils.functional import lazy
from adminsortable.fields import SortableForeignKey
from adminsortable.models import SortableMixin
from mptt.models import MPTTModel, TreeForeignKey
from cms.models import CMSPlugin
import uuid


class ProductType(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255, unique=True)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product type')
        verbose_name_plural = _('Product types')
        ordering = ['title']

    def __str__(self):
        return self.title

    def get_properties_fields(self, initial_values={}, for_variants=False):
        from django import forms
        if for_variants:
            properties = self.properties.filter(for_variants=True)
        else:
            properties = self.properties.all()
        result = [
            (
                'property_%s' % i,
                p.title,
                forms.CharField(label=p.title, initial=initial_values.get(p.title) or p.default_value, required=False),
            ) for i, p in enumerate(properties)]
        return result


class ProductTypeProperty(SortableMixin):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255)

    default_value = models.CharField(_('Default value'), max_length=255, blank=True)
    tag_prefix = models.CharField(_('Tag prefix'), max_length=10, blank=True)

    use_in_filters = models.BooleanField(_('Use in filters'), default=False)
    for_variants = models.BooleanField(_('For variants'), default=False)

    product_type = SortableForeignKey(ProductType, verbose_name=_('Product type'), related_name='properties')

    position = models.PositiveIntegerField(_('Position'), default=0, editable=False, db_index=True)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product type property')
        verbose_name_plural = _('Product type properties')
        ordering = ['position']

    def __str__(self):
        return self.title


# Listen for signals for ProductTypeProperty

@receiver(pre_save, sender=ProductTypeProperty)
def product_type_pre_save_listener(sender, instance, **kwargs):
    _resaved = False
    if instance.pk:
        from .tasks import rename_property_name, update_product_tags
        product_type_id = instance.product_type_id
        try:
            ptp = ProductTypeProperty.objects.get(pk=instance.pk)
            old_title = ptp.title
            new_title = instance.title
            if old_title != new_title:
                rename_property_name.delay(product_type_id, old_title, new_title)
                _resaved = True
            if not _resaved and ptp.tag_prefix != instance.tag_prefix:
                update_product_tags.delay(product_type_id, new_title)
        except ProductTypeProperty.DoesNotExist:
            pass


def generate_collection_image_filename(instance, filename):
    ext = filename.split('.')[-1]
    url = 'images/collections/%s/%s.%s' % (instance.id, str(uuid.uuid4()).replace('-', ''), ext)
    return url


class Collection(MPTTModel):

    DEFAULT_URL_FILTER_RULES = '''VENDOR
PRICE'''

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    site = models.ForeignKey(
        Site, verbose_name=_('Site'), related_name='collections', on_delete=models.PROTECT)
    title = models.CharField(_('Title'), max_length=255)
    slug = models.SlugField(_('Slug'))
    parent = TreeForeignKey(
        'self', null=True, blank=True, related_name='children', db_index=True,
        verbose_name=_('Parent'),
    )

    image = models.ImageField(_('Image'), null=True, blank=True, upload_to=generate_collection_image_filename)
    description = models.TextField(_('Description'), blank=True)

    page_title = models.CharField(_('Page title'), max_length=255, blank=True)
    meta_description = models.TextField(_('Meta description'), blank=True)

    url_filter_rules = models.TextField(
        _('URL filter rules'), blank=True, default=DEFAULT_URL_FILTER_RULES,
        help_text=_('Use separate lines for different rules. See documentation for details.')
    )

    show_vendor_filter = models.BooleanField(_('Show vendor filter'), default=True)
    show_properties_filters = models.BooleanField(_('Show properties filters'), default=True)
    exclude_properties_filters = ArrayField(
        models.CharField(max_length=70),
        verbose_name=_('Exclude properties filters'), default=list, blank=True,
        help_text=_('Comma separated list of properties labels which you want to exclude from filters.'),
    )
    show_price_filter = models.BooleanField(_('Show price filter'), default=True)

    published = models.BooleanField(_('Published'), default=True)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Collection')
        verbose_name_plural = _('Collections')

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        try:
            return reverse('pcart_collection:product-list-for-collection', args=[self.slug])
        except NoReverseMatch:
            return '#no-page-for-collection-app'

    def as_dict(self):
        result = {
            'id': self.id,
            'title': self.title,
            'site': {'domain': self.site.domain, 'name': self.site.name},
            'slug': self.slug,
            'parent': self.parent_id,
            'description': self.description,
            'page_title': self.page_title,
            'published': self.published,
            'added': self.added,
            'changed': self.changed,
            'url': self.get_absolute_url(),
        }
        return result


class ProductStatus(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255, unique=True)

    show_buy_button = models.BooleanField(_('Show buy button'), default=True)
    is_visible = models.BooleanField(_('Is visible'), default=True)
    is_searchable = models.BooleanField(_('Is searchable'), default=True)

    weight = models.PositiveIntegerField(_('Weight'), default=0)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product status')
        verbose_name_plural = _('Product statuses')
        ordering = ['id']  # TODO: change to weight

    def __str__(self):
        return self.title

    def as_dict(self):
        result = {
            'id': self.id,
            'title': self.title,
            'show_buy_button': self.show_buy_button,
            'is_visible': self.is_visible,
            'is_searchable': self.is_searchable,
            'weight': self.weight,
        }
        return result


def generate_vendor_image_filename(instance, filename):
    ext = filename.split('.')[-1]
    url = 'images/vendors/%s/%s.%s' % (instance.id, str(uuid.uuid4()).replace('-', ''), ext)
    return url


class Vendor(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255, unique=True)
    slug = models.SlugField(_('Slug'), max_length=255)
    image = models.ImageField(_('Image'), null=True, blank=True, upload_to=generate_vendor_image_filename)

    class Meta:
        verbose_name = _('Vendor')
        verbose_name_plural = _('Vendors')
        ordering = ['title']

    def __str__(self):
        return self.title

    def as_dict(self):
        result = {
            'id': self.id,
            'title': self.title,
        }
        return result


class Product(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255)
    slug = models.SlugField(_('Slug'), max_length=255)
    description = models.TextField(_('Description'), blank=True)

    page_title = models.CharField(_('Page title'), max_length=255, blank=True)
    meta_description = models.TextField(_('Meta description'), blank=True)

    vendor = models.ForeignKey(
        Vendor, verbose_name=_('Vendor'), related_name='products', null=True, blank=True, on_delete=models.SET_NULL)
    product_type = models.ForeignKey(
        ProductType, verbose_name=_('Product type'), related_name='products', on_delete=models.CASCADE)
    collections = models.ManyToManyField(
        Collection, verbose_name=_('Collections'), related_name='products', blank=True,
        help_text=_('Add this product to a collection so it\'s easy to find in your store.'),
    )

    tags = ArrayField(
        models.CharField(max_length=30),
        verbose_name=_('Tags'),
        blank=True,
        default=list,
    )
    properties = JSONField(_('Properties'), default=dict, blank=True)

    sku = models.CharField(_('SKU (Stock Keeping Unit)'), blank=True, max_length=100)
    barcode = models.CharField(_('Barcode (ISBN, UPC, GTIN, etc.)'), blank=True, max_length=100)
    status = models.ForeignKey(
        ProductStatus, verbose_name=_('Status'), related_name='products', on_delete=models.PROTECT)

    price = models.DecimalField(_('Price'), max_digits=10, decimal_places=2, default=0.00)
    compare_at_price = models.DecimalField(_('Compare at price'), max_digits=10, decimal_places=2, default=0.00)

    # Set automatically with stored procedure. For read only access.
    max_variant_price = models.DecimalField(_('Max variant price'), max_digits=10, decimal_places=2, default=0.00)
    min_variant_price = models.DecimalField(_('Min variant price'), max_digits=10, decimal_places=2, default=0.00)
    variants_count = models.PositiveIntegerField(_('Variants count'), default=0)

    weight = models.FloatField(
        _('Weight (kg)'), default=0.0, help_text=_('Used to calculate shipping rates at checkout.'))

    published = models.BooleanField(_('Published'), default=True)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product')
        verbose_name_plural = _('Products')
        ordering = ['id']

    def __str__(self):
        return self.title

    @staticmethod
    def type():
        return 'product'

    def get_absolute_url(self):
        try:
            return reverse('pcart_product:product-detail', args=[self.slug])
        except NoReverseMatch:
            return '#no-page-for-product-app'

    def get_properties_fields(self, initial_values=None, for_variants=False):
        return self.product_type.get_properties_fields(
            initial_values=self.properties if initial_values is None else initial_values,
            for_variants=for_variants)

    def has_variants(self):
        return self.variants_count > 0

    def get_image(self):
        return self.images.first()

    def available(self):
        return self.variants_count == 0 and self.price > 0 and self.status.show_buy_button

    def as_dict(self):
        result = {
            'id': self.id,
            'title': self.title,
            'slug': self.slug,
            'description': self.description,
            'page_title': self.page_title,
            'tags': self.tags,
            'properties': self.properties,
            'sku': self.sku,
            'barcode': self.barcode,
            'status': self.status.as_dict(),
            'price': self.price,
            'compare_at_price': self.compare_at_price,
            'max_variant_price': self.max_variant_price,
            'min_variant_price': self.min_variant_price,
            'variants_count': self.variants_count,
            'weight': self.weight,
            'published': self.published,
            'url': self.get_absolute_url(),
        }
        return result

    def get_page_title(self, request=None):
        return self.page_title or self.title


@receiver(post_save, sender=Product)
def product_post_save_listener(sender, instance, **kwargs):
    from .utils import filter_non_numeric_tags
    tagset = instance.tags
    values = filter_non_numeric_tags(tagset)
    PropertyValueSlug.objects.add_values(values)


class ProductVariant(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255)
    slug = models.SlugField(_('Slug'), unique=False, max_length=255)

    product = models.ForeignKey(
        Product, verbose_name=_('Product'), related_name='variants', on_delete=models.CASCADE)

    tags = ArrayField(
        models.CharField(max_length=30),
        verbose_name=_('Tags'),
        blank=True,
        default=list,
    )
    properties = JSONField(_('Properties'), default=dict, blank=True)

    sku = models.CharField(_('SKU (Stock Keeping Unit)'), blank=True, max_length=100)
    barcode = models.CharField(_('Barcode (ISBN, UPC, GTIN, etc.)'), blank=True, max_length=100)
    status = models.ForeignKey(ProductStatus, verbose_name=_('Status'), related_name='variants')

    price = models.DecimalField(_('Price'), max_digits=10, decimal_places=2, default=0.00)
    compare_at_price = models.DecimalField(_('Compare at price'), max_digits=10, decimal_places=2, default=0.00)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product variant')
        verbose_name_plural = _('Product variants')
        ordering = ['id']
        unique_together = ['product', 'slug']

    def __str__(self):
        return self.title

    @staticmethod
    def type():
        return 'variant'

    def get_absolute_url(self):
        try:
            return reverse('pcart_product:product-variant-detail', args=[self.product.slug, self.slug])
        except NoReverseMatch:
            return '#no-page-for-product-app'

    def get_properties_fields(self, initial_values=None):
        return self.product.product_type.get_properties_fields(
            initial_values=self.properties if initial_values is None else initial_values,
            for_variants=True)

    def get_image(self):
        return self.product.get_image()

    def weight(self):
        return self.product.weight

    def available(self):
        return self.price > 0 and self.status.show_buy_button


def generate_product_image_filename(instance, filename):
    ext = filename.split('.')[-1]
    url = 'images/products/%s/%s.%s' % (instance.product.id, str(uuid.uuid4()).replace('-', ''), ext)
    return url


class ProductImage(SortableMixin):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255, default='', blank=True)
    product = SortableForeignKey(
        Product, verbose_name=_('Product'), related_name='images', on_delete=models.CASCADE)

    image = models.ImageField(_('Image'), null=True, blank=True, upload_to=generate_product_image_filename)
    html_snippet = models.TextField(_('HTML snippet'), default='', blank=True)

    tags = ArrayField(
        models.CharField(max_length=30),
        verbose_name=_('Tags'),
        blank=True,
        default=list,
    )

    download_link = models.CharField(_('Download link'), max_length=300, default='', blank=True)
    downloaded = models.BooleanField(_('Downloaded'), default=False)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    position = models.PositiveIntegerField(_('Position'), default=0, editable=False, db_index=True)

    class Meta:
        verbose_name = _('Product image')
        verbose_name_plural = _('Product images')
        ordering = ['position']

    def __str__(self):
        return self.title or self.product.title

    def download_product_image(self, save=False):
        from django.core.files import File
        from ftplib import FTP
        from io import BytesIO
        import urllib.request
        import urllib.parse
        import os
        _changed = False
        o = urllib.parse.urlparse(self.download_link)

        if self.downloaded is False:
            if o.scheme in ['http', 'https']:
                # Download an image via HTTP
                local_filename, headers = urllib.request.urlretrieve(self.download_link)
                self.image.save(
                    os.path.basename(o.path),
                    File(open(local_filename, 'rb')),
                    save=False,
                )
                self.downloaded = True
                _changed = True
            elif o.scheme == 'ftp':
                # Download an image via FTP
                _ftp_sources = getattr(settings, 'PCART_FTP_IMAGE_SOURCES', {})
                _credentials = _ftp_sources.get(o.netloc)
                with FTP(o.netloc) as ftp:
                    if _credentials:
                        ftp.login(_credentials.get('user'), _credentials.get('password'))
                    else:
                        ftp.login()
                    ftp.getwelcome()
                    r = BytesIO()
                    ftp.retrbinary('RETR ~/%s' % o.path, r.write)
                    r.seek(0)

                    self.image.save(
                        os.path.basename(o.path),
                        File(r),
                        save=False,
                    )
                    self.downloaded = True
                    _changed = True

        if save and _changed:
            self.save()

    def save(self, *args, **kwargs):
        if self.downloaded is False and self.download_link:
            self.download_product_image(save=False)
        super(ProductImage, self).save(*args, **kwargs)


@receiver(pre_save, sender=ProductImage)
def product_image_pre_save_listener(sender, instance, **kwargs):
    if instance.pk:
        try:
            p_im = ProductImage.objects.get(pk=instance.pk)
            if p_im.image and p_im.image != instance.image:
                p_im.image.delete(save=False)
        except ProductImage.DoesNotExist:
            pass


@receiver(post_delete, sender=ProductImage)
def product_image_post_delete_listener(sender, instance, **kwargs):
    if instance.image:
        instance.image.delete(save=False)


class PropertyValueSlugManager(models.Manager):
    def generate_unique_slug(self, value):
        from pcart_core.utils import get_unique_slug
        slug = get_unique_slug(value, self.model)
        return slug

    def add_values(self, values=[]):
        values = list(set(values))
        existing_values = self.filter(value__in=values).values_list('value', flat=True)
        for v in values:
            if v not in existing_values:
                slug = self.generate_unique_slug(v)
                item = self.model(value=v, slug=slug)
                item.save(using=self._db)

    def update_data(self, products=None):
        from .utils import get_unique_tagset, filter_non_numeric_tags
        values = filter_non_numeric_tags(get_unique_tagset(products))
        self.add_values(values)


class PropertyValueSlug(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    value = models.CharField(_('Value'), max_length=255, unique=True)
    slug = models.SlugField(_('Slug'), max_length=255)

    objects = PropertyValueSlugManager()

    class Meta:
        verbose_name = _('Property value slug')
        verbose_name_plural = _('Property values slugs')
        ordering = ['value']

    def __str__(self):
        return self.value


# DjangoCMS plugins


class CollectionPluginModel(CMSPlugin):
    """ Represents a plugin with a products from the specified collection.
    """
    collection = TreeForeignKey(Collection, verbose_name=_('Collection'))
    filter_string = models.CharField(
        _('Filter string'), default='', blank=True, max_length=255,
        help_text=_('Use the list of filter chunks separated with / and ; characters.')
    )
    sorting = models.CharField(_('Sorting'), default='', blank=True, max_length=100)
    template_name = models.CharField(_('Template name'), default='', blank=True, max_length=100)

    def __init__(self, *args, **kwargs):
        super(CollectionPluginModel, self).__init__(*args, **kwargs)
        # self._meta.get_field_by_name('sorting')[0]._choices = get_plugin_sorting_choices()
        # self._meta.get_field_by_name('template_name')[0]._choices = get_plugin_template_choices()

    def __str__(self):
        return str(self.collection)



from celery import task
# from django.conf import settings


@task
def rename_property_name(product_type_id, old_name, new_name):
    if old_name == new_name:
        return None
    from .models import ProductType, Product, ProductVariant
    product_type = ProductType.objects.get(pk=product_type_id)
    products = Product.objects.filter(
        product_type=product_type, properties__has_key=old_name)
    product_variants = ProductVariant.objects.filter(
        product__product_type=product_type, properties__has_key=old_name)

    for p in products:
        _props = p.properties
        if old_name in _props:
            _props[new_name] = _props[old_name]
            del _props[old_name]
            p.save()

    for p in product_variants:
        _props = p.properties
        if old_name in _props:
            _props[new_name] = _props[old_name]
            del _props[old_name]
            p.save()


@task
def update_product_tags(product_type_id, property_title):
    from django.db.models import F
    from .models import ProductType, Product, ProductVariant
    product_type = ProductType.objects.get(pk=product_type_id)
    products = Product.objects.filter(
        product_type=product_type, properties__has_key=property_title)
    product_variants = ProductVariant.objects.filter(
        product__product_type=product_type, properties__has_key=property_title)

    products.update(sku=F('sku'))
    product_variants.update(sku=F('sku'))


# Run this task automatically every 6 hours
@task.periodic_task(run_every=60*60*6)
def update_property_value_slugs():
    from .models import PropertyValueSlug
    PropertyValueSlug.objects.update_data()



from django.utils.translation import ugettext as _
from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from .models import CollectionPluginModel
from .forms import CollectionPluginForm


class CollectionPluginPublisher(CMSPluginBase):
    model = CollectionPluginModel  # model where plugin data are saved
    form = CollectionPluginForm
    module = _("Catalog")
    name = _("Collection")  # name of the plugin in the interface
    render_template = "catalog/plugins/collection_plugin.html"

    def render(self, context, instance, placeholder):
        from .filtering import ProductsFilterManager
        from .utils import filter_slug_to_tags

        filter_manager = ProductsFilterManager(
            collection=instance.collection,
            view=instance.template_name,
            sort=instance.sorting)
        filter_chunks = instance.filter_string.split('/')
        filter_tags, vendors, prices, normalized_url_chunks, _redirect = filter_slug_to_tags(
            instance.collection, filter_chunks)
        filter_manager.set_filters(filter_tags=filter_tags, vendors=vendors, prices=prices)
        context = filter_manager.get_context()
        context.update({'instance': instance})
        return context

plugin_pool.register_plugin(CollectionPluginPublisher)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-20 18:32
from __future__ import unicode_literals

from django.db import migrations, models
import pcart_catalog.models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0003_auto_20170209_2135'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='producttype',
            options={'ordering': ['title'], 'verbose_name': 'Product type', 'verbose_name_plural': 'Product types'},
        ),
        migrations.AlterModelOptions(
            name='producttypeproperty',
            options={'ordering': ['position'], 'verbose_name': 'Product type property', 'verbose_name_plural': 'Product type properties'},
        ),
        migrations.AddField(
            model_name='collection',
            name='image',
            field=models.ImageField(blank=True, null=True, upload_to=pcart_catalog.models.generate_collection_image_filename, verbose_name='Image'),
        ),
        migrations.AddField(
            model_name='producttypeproperty',
            name='position',
            field=models.PositiveIntegerField(default=0, verbose_name='Position'),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='compare_at_price',
            field=models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Compare at price'),
        ),
        migrations.AddField(
            model_name='vendor',
            name='image',
            field=models.ImageField(blank=True, null=True, upload_to=pcart_catalog.models.generate_vendor_image_filename, verbose_name='Image'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-09 19:30
from __future__ import unicode_literals

import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import mptt.fields
import pcart_catalog.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
    ]

    operations = [
        migrations.CreateModel(
            name='Collection',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('slug', models.SlugField(verbose_name='Slug')),
                ('description', models.TextField(blank=True, verbose_name='Description')),
                ('page_title', models.CharField(blank=True, max_length=255, verbose_name='Page title')),
                ('meta_description', models.TextField(blank=True, verbose_name='Meta description')),
                ('url_filter_rules', models.TextField(blank=True, default='', help_text='Use separate lines for different rules. See documentation for details.', verbose_name='URL filter rules')),
                ('published', models.BooleanField(default=True, verbose_name='Published')),
                ('added', models.DateTimeField(auto_now_add=True, verbose_name='Added')),
                ('changed', models.DateTimeField(auto_now=True, verbose_name='Changed')),
                ('lft', models.PositiveIntegerField(db_index=True, editable=False)),
                ('rght', models.PositiveIntegerField(db_index=True, editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(db_index=True, editable=False)),
                ('parent', mptt.fields.TreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='pcart_catalog.Collection', verbose_name='Parent')),
                ('site', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='collections', to='sites.Site', verbose_name='Site')),
            ],
            options={
                'verbose_name': 'Collection',
                'verbose_name_plural': 'Collections',
            },
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('slug', models.SlugField(max_length=255, verbose_name='Slug')),
                ('description', models.TextField(blank=True, verbose_name='Description')),
                ('page_title', models.CharField(blank=True, max_length=255, verbose_name='Page title')),
                ('meta_description', models.TextField(blank=True, verbose_name='Meta description')),
                ('tags', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=30), blank=True, default=list, size=None, verbose_name='Tags')),
                ('properties', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, verbose_name='Properties')),
                ('sku', models.CharField(blank=True, max_length=100, verbose_name='SKU (Stock Keeping Unit)')),
                ('barcode', models.CharField(blank=True, max_length=100, verbose_name='Barcode (ISBN, UPC, GTIN, etc.)')),
                ('price', models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Price')),
                ('compare_at_price', models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Compare at price')),
                ('max_variant_price', models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Max variant price')),
                ('min_variant_price', models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Min variant price')),
                ('variants_count', models.PositiveIntegerField(default=0, verbose_name='Variants count')),
                ('weight', models.FloatField(default=0.0, help_text='Used to calculate shipping rates at checkout.', verbose_name='Weight (kg)')),
                ('published', models.BooleanField(default=True, verbose_name='Published')),
                ('added', models.DateTimeField(auto_now_add=True, verbose_name='Added')),
                ('changed', models.DateTimeField(auto_now=True, verbose_name='Changed')),
                ('collections', models.ManyToManyField(blank=True, help_text="Add this product to a collection so it's easy to find in your store.", related_name='products', to='pcart_catalog.Collection', verbose_name='Collections')),
            ],
            options={
                'verbose_name': 'Product',
                'verbose_name_plural': 'Products',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='ProductImage',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(blank=True, default='', max_length=255, verbose_name='Title')),
                ('image', models.ImageField(blank=True, null=True, upload_to=pcart_catalog.models.generate_product_image_filename, verbose_name='Image')),
                ('html_snippet', models.TextField(blank=True, default='', verbose_name='HTML snippet')),
                ('tags', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=30), blank=True, default=list, size=None, verbose_name='Tags')),
                ('position', models.PositiveIntegerField(default=0, verbose_name='Position')),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images', to='pcart_catalog.Product', verbose_name='Product')),
            ],
            options={
                'verbose_name': 'Product image',
                'verbose_name_plural': 'Product images',
            },
        ),
        migrations.CreateModel(
            name='ProductStatus',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255, unique=True, verbose_name='Title')),
                ('show_buy_button', models.BooleanField(default=True, verbose_name='Show buy button')),
                ('is_visible', models.BooleanField(default=True, verbose_name='Is visible')),
                ('is_searchable', models.BooleanField(default=True, verbose_name='Is searchable')),
                ('weight', models.PositiveIntegerField(default=0, verbose_name='Weight')),
                ('added', models.DateTimeField(auto_now_add=True, verbose_name='Added')),
                ('changed', models.DateTimeField(auto_now=True, verbose_name='Changed')),
            ],
            options={
                'verbose_name': 'Product status',
                'verbose_name_plural': 'Product statuses',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='ProductType',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255, unique=True, verbose_name='Title')),
                ('added', models.DateTimeField(auto_now_add=True, verbose_name='Added')),
                ('changed', models.DateTimeField(auto_now=True, verbose_name='Changed')),
            ],
            options={
                'verbose_name': 'Product type',
                'verbose_name_plural': 'Product types',
                'ordering': ['title', 'id'],
            },
        ),
        migrations.CreateModel(
            name='ProductTypeProperty',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('default_value', models.CharField(blank=True, max_length=255, verbose_name='Default value')),
                ('tag_prefix', models.CharField(blank=True, max_length=10, verbose_name='Tag prefix')),
                ('use_in_filters', models.BooleanField(default=False, verbose_name='Use in filters')),
                ('for_variants', models.BooleanField(default=False, verbose_name='For variants')),
                ('added', models.DateTimeField(auto_now_add=True, verbose_name='Added')),
                ('changed', models.DateTimeField(auto_now=True, verbose_name='Changed')),
                ('product_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='pcart_catalog.ProductType', verbose_name='Product type')),
            ],
            options={
                'verbose_name': 'Product type property',
                'verbose_name_plural': 'Product type properties',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='ProductVariant',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('slug', models.SlugField(max_length=255, verbose_name='Slug')),
                ('tags', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=30), blank=True, default=list, size=None, verbose_name='Tags')),
                ('properties', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, verbose_name='Properties')),
                ('sku', models.CharField(blank=True, max_length=100, verbose_name='SKU (Stock Keeping Unit)')),
                ('barcode', models.CharField(blank=True, max_length=100, verbose_name='Barcode (ISBN, UPC, GTIN, etc.)')),
                ('price', models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Price')),
                ('added', models.DateTimeField(auto_now_add=True, verbose_name='Added')),
                ('changed', models.DateTimeField(auto_now=True, verbose_name='Changed')),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variants', to='pcart_catalog.Product', verbose_name='Product')),
                ('status', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variants', to='pcart_catalog.ProductStatus', verbose_name='Status')),
            ],
            options={
                'verbose_name': 'Product variant',
                'verbose_name_plural': 'Product variants',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='Vendor',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255, unique=True, verbose_name='Title')),
            ],
            options={
                'verbose_name': 'Vendor',
                'verbose_name_plural': 'Vendors',
                'ordering': ['title'],
            },
        ),
        migrations.AddField(
            model_name='product',
            name='product_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='products', to='pcart_catalog.ProductType', verbose_name='Product type'),
        ),
        migrations.AddField(
            model_name='product',
            name='status',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='products', to='pcart_catalog.ProductStatus', verbose_name='Status'),
        ),
        migrations.AddField(
            model_name='product',
            name='vendor',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='products', to='pcart_catalog.Vendor', verbose_name='Vendor'),
        ),
        migrations.AlterUniqueTogether(
            name='productvariant',
            unique_together=set([('product', 'slug')]),
        ),
    ]



from django.conf.urls import url
from . import views


urlpatterns = [
    url(
        r'^$', views.collections_list_view, name='all-collections'),
    url(
        r'^(?P<collection_slug>[\d\w\-]+)/filter-form-dispatcher/$',
        views.filter_form_dispatcher,
        name='filter-form-dispatcher'),
    url(
        r'^(?P<slug>.+)/$',
        views.collection_view,
        name='product-list-for-collection'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-09 09:43
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0005_auto_20170220_2047'),
    ]

    operations = [
        migrations.AddField(
            model_name='productimage',
            name='download_link',
            field=models.CharField(blank=True, default='', max_length=300, verbose_name='Download link'),
        ),
    ]



from django.contrib import admin
from mptt.admin import MPTTModelAdmin
from django.utils.translation import ugettext_lazy as _
from adminsortable.admin import NonSortableParentAdmin, SortableTabularInline
from django.utils.safestring import mark_safe
from cms.admin.placeholderadmin import FrontendEditableAdminMixin
from pcart_treeadmin.admin import TreeAdmin
from .models import (
    Collection,
    ProductType,
    ProductTypeProperty,
    ProductStatus,
    Product,
    ProductVariant,
    ProductImage,
    Vendor,
    PropertyValueSlug,
)
from .forms import EditProductVariantForm, EditProductVariantFormSet


class VendorAdmin(admin.ModelAdmin):
    list_display = ('title', 'slug', 'id')
    search_fields = ['title', 'slug', 'id']
    prepopulated_fields = {"slug": ("title",)}

admin.site.register(Vendor, VendorAdmin)


class CollectionAdmin(TreeAdmin):
    prepopulated_fields = {"slug": ("title",)}
    list_display = ('title_preview', 'id', 'changed', 'published', 'site', 'slug')
    list_filter = ('published', 'site')
    preserve_filters = True
    search_fields = ['title', 'id']
    fieldsets = (
        (None, {
            'fields': ('site', 'title', 'slug', 'parent', 'description', 'image'),
        }),
        (_('SEO'), {
            'fields': ('page_title', 'meta_description'),
            'classes': ('collapse',)
        }),
        (_('Filters'), {
            'fields': (
                'url_filter_rules',
                'show_vendor_filter', 'show_properties_filters', 'exclude_properties_filters',
                'show_price_filter',
            ),
        }),
        (_('Publication'), {
            'fields': ('published',),
        }),
    )

    def title_preview(self, obj):
        return mark_safe('%s%s' % ('&nbsp;'*obj.level*4, obj.title))
    title_preview.short_description = _('Title')
    title_preview.admin_order_field = 'title'

admin.site.register(Collection, CollectionAdmin)


class ProductTypePropertyInline(SortableTabularInline):
    model = ProductTypeProperty
    extra = 1


class ProductTypeAdmin(NonSortableParentAdmin):
    list_display = ('title', 'id')
    search_fields = ['title']
    inlines = [ProductTypePropertyInline]

admin.site.register(ProductType, ProductTypeAdmin)


class ProductStatusAdmin(admin.ModelAdmin):
    list_display = ('title', 'id', 'show_buy_button', 'is_visible', 'is_searchable', 'weight')
    search_fields = ['title']

admin.site.register(ProductStatus, ProductStatusAdmin)


class ProductImageInline(SortableTabularInline):
    model = ProductImage
    extra = 1


class ProductVariantInline(admin.TabularInline):
    model = ProductVariant
    readonly_fields = ['tags']
    form = EditProductVariantForm
    formset = EditProductVariantFormSet
    extra = 1

    def get_fields(self, request, obj=None):
        exclude_prop_fields = tuple(p[0] for p in obj.get_properties_fields(for_variants=True))
        self.exclude = exclude_prop_fields
        fields = ('title', 'slug', 'tags', 'sku', 'barcode', 'status', 'price', 'compare_at_price') + exclude_prop_fields
        return fields


class ProductAdmin(FrontendEditableAdminMixin, NonSortableParentAdmin):
    prepopulated_fields = {"slug": ("title",)}
    list_display = (
        'show_title',
        'vendor',
        'status',
        'show_price',
        'product_type',
        'get_collections',
        'published', 'slug')
    list_filter = ('published', 'collections', 'product_type')
    filter_horizontal = ('collections',)
    preserve_filters = True
    search_fields = ['title', 'id']
    inlines = [ProductImageInline, ProductVariantInline]
    save_as_continue = True

    def get_queryset(self, request):
        queryset = super(ProductAdmin, self).get_queryset(request)
        queryset = queryset\
            .prefetch_related('vendor')\
            .prefetch_related('status') \
            .prefetch_related('product_type') \
            .prefetch_related('collections')
        return queryset

    def show_price(self, obj):
        if obj.variants_count > 0:
            return '%s - %s' % (obj.min_variant_price, obj.max_variant_price)
        else:
            if obj.compare_at_price != 0:
                return '%s --%s--' % (obj.price, obj.compare_at_price)
            else:
                return '%s' % obj.price
    show_price.short_description = _('Price')
    show_price.admin_order_field = 'min_variant_price'

    def get_collections(self, obj):
        return ', '.join([x.title for x in obj.collections.all()])
    get_collections.short_description = _('Collections')

    def show_title(self, obj):
        from django.utils.html import format_html
        if obj.variants_count > 0:
            return format_html(
                '{0} <span style="float:right;"><strong>({1})</strong></span>',
                obj.title,
                obj.variants_count)
        else:
            return obj.title
    show_title.short_description = _('Title')
    show_title.admin_order_field = 'title'

    def get_form(self, request, obj=None, **kwargs):
        from .forms import AddProductForm, EditProductForm
        if obj is None:
            kwargs['form'] = AddProductForm
        else:
            kwargs['form'] = EditProductForm
        return super(ProductAdmin, self).get_form(request, obj, **kwargs)

    def get_fieldsets(self, request, obj=None):
        if obj is None:
            self.exclude = []
            return (
                (None, {
                    'fields': ('title', 'slug', 'product_type', 'vendor', 'status'),
                }),
            )
        else:
            exclude_prop_fields = tuple(p[0] for p in obj.get_properties_fields())
            self.exclude = exclude_prop_fields
            return (
                (None, {
                    'fields': (
                        'title', 'slug', 'product_type', 'vendor', 'description'),
                }),
                (_('SEO'), {
                    'fields': ('page_title', 'meta_description'),
                    'classes': ('collapse',),
                }),
                (_('Tags'), {
                    'fields': ('tags', 'sku', 'barcode', 'status'),
                }),
                (_('Properties'), {
                    'fields': exclude_prop_fields,
                }),
                (_('Price'), {
                    'fields': ('price', 'compare_at_price', 'max_variant_price', 'min_variant_price'),
                }),
                (_('Shipping'), {
                    'fields': ('weight',),
                    'classes': ('collapse',),
                }),
                (_('Publication'), {
                    'fields': ('collections', 'published', 'variants_count', 'added', 'changed'),
                }),
            )

    def get_formsets_with_inlines(self, request, obj=None):
        for inline in self.get_inline_instances(request, obj):
            # hide ProductVariantInline in the add view
            if isinstance(inline, ProductVariantInline) and obj is None:
                continue
            yield inline.get_formset(request, obj), inline

    def get_readonly_fields(self, request, obj=None):
        if obj is None:
            return []
        else:
            if obj.variants_count == 0:
                return (
                    'product_type',
                    'tags',
                    'min_variant_price',
                    'max_variant_price',
                    'variants_count',
                    'added',
                    'changed',
                )
            else:
                return (
                    'product_type',
                    'tags',
                    'status',
                    'min_variant_price',
                    'max_variant_price',
                    'variants_count',
                    'added',
                    'changed',
                )

admin.site.register(Product, ProductAdmin)


class PropertyValueSlugAdmin(admin.ModelAdmin):
    list_display = ('value', 'slug')
    search_fields = ('value', 'slug')

admin.site.register(PropertyValueSlug, PropertyValueSlugAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-04-25 10:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0014_auto_20170411_1025'),
    ]

    operations = [
        migrations.AddField(
            model_name='vendor',
            name='slug',
            field=models.SlugField(max_length=255, null=True, verbose_name='Slug'),
        ),
    ]



from cms.wizards.wizard_base import Wizard
from cms.wizards.wizard_pool import wizard_pool
from django.utils.translation import ugettext_lazy as _
from .forms import AddProductForm
from .models import Product


class ProductWizard(Wizard):
    pass

product_wizard = ProductWizard(
    title=_('New product'),
    weight=200,
    form=AddProductForm,
    description=_('Create a new product'),
)

wizard_pool.register(product_wizard)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-09 09:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0006_productimage_download_link'),
    ]

    operations = [
        migrations.AddField(
            model_name='productimage',
            name='added',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='Added'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='productimage',
            name='changed',
            field=models.DateTimeField(auto_now=True, verbose_name='Changed'),
        ),
    ]



from django import forms
from django.utils.translation import ugettext_lazy as _
from django.forms import BaseInlineFormSet
from .models import Product, ProductVariant, CollectionPluginModel


def get_plugin_sorting_choices():
    from .settings import PCART_COLLECTION_ORDERINGS
    result = []
    for k in PCART_COLLECTION_ORDERINGS.keys():
        result.append((k, PCART_COLLECTION_ORDERINGS[k].get('label')))
    return result


def get_plugin_template_choices():
    from django.conf import settings
    result = []
    for k in settings.PCART_COLLECTION_TEMPLATES:
        if settings.PCART_COLLECTION_TEMPLATES[k].get('plugins_only', False):
            result.append((
                k,
                settings.PCART_COLLECTION_TEMPLATES[k].get('label') or
                settings.PCART_COLLECTION_TEMPLATES[k].get('template')
            ))
    return result


class AddProductForm(forms.ModelForm):
    class Meta:
        model = Product
        fields = ['title', 'slug', 'product_type', 'status']


class EditProductForm(forms.ModelForm):
    class Meta:
        model = Product
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super(EditProductForm, self).__init__(*args, **kwargs)
        if 'instance' in kwargs:
            obj = kwargs['instance']
            properties_fields = obj.get_properties_fields()
            for p in properties_fields:
                self.fields[p[0]] = p[2]

    def save(self, commit=True):
        instance = super(EditProductForm, self).save(commit=False)
        properties = {}
        properties_fields = instance.get_properties_fields()
        for p in properties_fields:
            _value = self.cleaned_data[p[0]]
            if _value:
                properties[p[1]] = _value
        instance.properties = properties
        if commit:
            instance.save()
        return instance


class EditProductVariantForm(forms.ModelForm):
    class Meta:
        model = ProductVariant
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        self.product_instance = kwargs.pop('product_instance')
        super(EditProductVariantForm, self).__init__(*args, **kwargs)
        initial_values = {}
        if 'instance' in kwargs:
            initial_values = kwargs['instance'].properties
        properties_fields = self.product_instance.get_properties_fields(initial_values=initial_values, for_variants=True)
        for p in properties_fields:
            self.fields[p[0]] = p[2]

    def save(self, commit=True):
        instance = super(EditProductVariantForm, self).save(commit=False)
        properties = self.product_instance.properties
        properties_fields = instance.get_properties_fields()
        for p in properties_fields:
            _value = self.cleaned_data[p[0]]
            if _value:
                properties[p[1]] = _value

        instance.properties = properties
        if commit:
            instance.save()
        return instance


class EditProductVariantFormSet(BaseInlineFormSet):
    def __init__(self, *args, **kwargs):
        super(EditProductVariantFormSet, self).__init__(*args, **kwargs)
        self.form_kwargs['product_instance'] = kwargs['instance']


class CollectionPluginForm(forms.ModelForm):
    model = CollectionPluginModel

    class Meta:
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super(CollectionPluginForm, self).__init__(*args, **kwargs)
        self.fields['sorting'] = forms.ChoiceField(
            choices=get_plugin_sorting_choices(),
            label=_('Sorting'),
        )
        self.fields['template_name'] = forms.ChoiceField(
            choices=get_plugin_template_choices(),
            label=_('Template name'),
        )



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from cms.menu_bases import CMSAttachMenu
from menus.base import Menu, NavigationNode
from menus.menu_pool import menu_pool


class CollectionMenu(CMSAttachMenu):
    name = _("Collection menu")

    def get_nodes(self, request):
        """
        This method is used to build the menu tree.
        """
        from .models import Collection
        nodes = []
        for collection in Collection.objects.filter(published=True):
            node = NavigationNode(
                collection.title,
                collection.get_absolute_url(),
                collection.slug
            )
            nodes.append(node)
        return nodes

menu_pool.register_menu(CollectionMenu)


from django.utils.translation import ugettext_lazy as _
from cms.toolbar_pool import toolbar_pool
from cms.toolbar_base import CMSToolbar
from cms.utils.urlutils import admin_reverse
from .models import Product, Collection


@toolbar_pool.register
class CollectionToolbar(CMSToolbar):
    supported_apps = (
        'pcart_catalog',
    )
    watch_models = [Collection,]

    def populate(self):
        if not self.is_current_app:
            return

        menu = self.toolbar.get_or_create_menu('collection-app', _('Collection'))

        menu.add_sideframe_item(
            name=_('Collections list'),
            url=admin_reverse('pcart_catalog_collection_changelist'),
        )

        menu.add_modal_item(
            name=_('Add new collection'),
            url=admin_reverse('pcart_catalog_collection_add'),
        )
        # menu.add_modal_item(
        #     name=_('Edit collection'),
        #     url=admin_reverse('pcart_catalog_collection_change', [self.]),
        # )


@toolbar_pool.register
class ProductToolbar(CMSToolbar):
    supported_apps = (
        'pcart_catalog',
    )
    watch_models = [Product,]

    def populate(self):
        if not self.is_current_app:
            return

        menu = self.toolbar.get_or_create_menu('product-app', _('Product'))

        menu.add_sideframe_item(
            name=_('Products list'),
            url=admin_reverse('pcart_catalog_product_changelist'),
        )

        current_product_id = getattr(self.request, 'productid', None)
        if current_product_id:
            menu.add_modal_item(
                name=_('Edit product'),
                url=admin_reverse('pcart_catalog_product_change', args=(current_product_id,)),
                active=True,
            )

        menu.add_modal_item(
            name=_('Add new product'),
            url=admin_reverse('pcart_catalog_product_add'),
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-05-05 07:04
from __future__ import unicode_literals

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0017_auto_20170425_1325'),
    ]

    operations = [
        migrations.CreateModel(
            name='PropertyValueSlug',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('value', models.CharField(max_length=255, unique=True, verbose_name='Value')),
                ('slug', models.SlugField(max_length=255, verbose_name='Slug')),
            ],
            options={
                'ordering': ['value'],
                'verbose_name_plural': 'Property values slugs',
                'verbose_name': 'Property value slug',
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-09 19:35
from __future__ import unicode_literals

from django.db import migrations

SQL = '''
CREATE INDEX idx_product_tags ON "pcart_catalog_product" USING GIN("tags");
CREATE INDEX idx_productvariant_tags ON "pcart_catalog_productvariant" USING GIN("tags");
'''

class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0002_auto_20170209_2130'),
    ]

    operations = [
        migrations.RunSQL(SQL),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-26 16:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0010_auto_20170326_1953'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='exclude_properties_filters',
            field=models.CharField(blank=True, default='', help_text='Comma separated list of properties tags which you want to exclude from filters.', max_length=255, verbose_name='Exclude properties filters'),
        ),
    ]



from django.conf.urls import url
from . import views


urlpatterns = [
    url(r'^$', views.redirect_to_collections, name='all-products'),
    url(r'^(?P<product_slug>[-\w]+)/$', views.ProductDetailView.as_view(), name='product-detail'),
    url(
        r'^(?P<product_slug>[-\w]+)/(?P<variant_slug>[-\w]+)/$',
        views.ProductVariantDetailView.as_view(), name='product-variant-detail'),
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.contrib.staticfiles import finders
try:
    from PIL import Image, ImageEnhance
except ImportError:
    import Image, ImageEnhance

"""
Utilities for applying a watermark to an image using PIL.

Original Source: http://code.activestate.com/recipes/362879/
                 https://github.com/codekoala/django-watermark/blob/master/watermarker/utils.py
"""

import os
import re
import random

def _percent(var):
    """
    Just a simple interface to the _val function with a more meaningful name.
    """
    return _val(var, True)

def _int(var):
    """
    Just a simple interface to the _val function with a more meaningful name.
    """
    return _val(var)

def _val(var, is_percent=False):
    """
    Tries to determine the appropriate value of a particular variable that is
    passed in. If the value is supposed to be a percentage, a whole integer
    will be sought after and then turned into a floating point number between
    0 and 1. If the value is supposed to be an integer, the variable is cast
    into an integer.
    """
    try:
        if is_percent:
            var = float(int(var.strip('%')) / 100.0)
        else:
            var = int(var)
    except ValueError:
        raise ValueError('invalid watermark parameter: ' + var)
    return var

def reduce_opacity(img, opacity):
    """
    Returns an image with reduced opacity.
    """
    assert opacity >= 0 and opacity <= 1

    if img.mode != 'RGBA':
        img = img.convert('RGBA')
    else:
        img = img.copy()

    alpha = img.split()[3]
    alpha = ImageEnhance.Brightness(alpha).enhance(opacity)
    img.putalpha(alpha)

    return img

def determine_scale(scale, img, mark):
    """
    Scales an image using a specified ratio or 'F'. If `scale` is 'F', the
    image is scaled to be as big as possible to fit in `img` without falling off
    the edges. Returns the scaled `mark`.
    """

    if scale:
        print('scale=', scale)
        try:
            scale = float(scale)
        except (ValueError, TypeError):
            pass

        errmess = 'Invalid scale value "%s"! Valid values are 1) "F" for ratio-preserving scaling and 2) floating-point numbers and integers greater than 0.' % (scale,)
        if type(scale) is str:
            if scale.lower() == 'f':
                # scale, but preserve the aspect ratio
                scale = min(
                            float(img.size[0]) / mark.size[0],
                            float(img.size[1]) / mark.size[1]
                           )
            elif re.match(r'\d{1,3}\%', scale):
                scale = float(scale.replace('%',''))
                scale = min(
                            float(img.size[0]) / mark.size[0],
                            float(img.size[1]) / mark.size[1]
                           ) / 100 * scale

            else:
                raise ValueError(errmess)
        elif type(scale) not in (float, int):
            raise ValueError(errmess)

        # determine the new width and height
        w = int(mark.size[0] * float(scale))
        h = int(mark.size[1] * float(scale))

        # apply the new width and height, and return the new `mark`
        return (w, h)
    else:
        return mark.size

def determine_rotation(rotation, mark):
    """
    Determines the number of degrees to rotate the watermark image.
    """

    if isinstance(rotation, str) \
        and rotation.lower() == 'r':
        rotation = random.randint(0, 359)
    else:
        rotation = _int(rotation)

    return rotation

def determine_position(position, img, mark):
    """
    Options:
    TL: top-left
    TR: top-right
    BR: bottom-right
    BL: bottom-left
    C: centered
    R: random
    X%xY%: relative positioning on both the X and Y axes
    X%xY: relative positioning on the X axis and absolute positioning on the
    Y axis
    XxY%: absolute positioning on the X axis and relative positioning on the
    Y axis
    XxY: absolute positioning on both the X and Y axes
    """

    max_left = max(img.size[0] - mark.size[0], 0)
    max_top = max(img.size[1] - mark.size[1], 0)

    if not position:
        position = 'r'

    if isinstance(position, tuple):
        left, top = position
    elif isinstance(position, str):
        position = position.lower()

        # corner positioning
        if position in ['tl', 'tr', 'br', 'bl']:
            if 't' in position:
                top = 0
            elif 'b' in position:
                top = max_top
            if 'l' in position:
                left = 0
            elif 'r' in position:
                left = max_left

        # center positioning
        elif position == 'c':
            left = int(max_left / 2)
            top = int(max_top / 2)

        # random positioning
        elif position == 'r':
            left = random.randint(0, max_left)
            top = random.randint(0, max_top)

        # relative or absolute positioning
        elif 'x' in position:
            left, top = position.split('x')

            if '%' in left:
                left = max_left * _percent(left)
            else:
                left = _int(left)

            if '%' in top:
                top = max_top * _percent(top)
            else:
                top = _int(top)

    return (int(left), int(top))

def watermark_filter(img, mark, position, opacity, scale, tile, greyscale, rotation, **kwargs):
    """
    Adds a watermark to an image.
    """
    if opacity < 1:
        mark = reduce_opacity(mark, opacity)

    if type(scale) != tuple:
        scale = determine_scale(scale, img, mark)

    if scale[0] != mark.size[0] and scale[1] != mark.size[1]:
        mark = mark.resize(scale)

    if greyscale and mark.mode != 'LA':
        mark = mark.convert('LA')

    rotation = determine_rotation(rotation, mark)
    if rotation != 0:
        # give some leeway for rotation overlapping
        new_w = mark.size[0] * 1.5
        new_h = mark.size[1] * 1.5

        new_mark = Image.new('RGBA', (new_w, new_h), (0,0,0,0))

        # center the watermark in the newly resized image
        new_l = (new_w - mark.size[0]) / 2
        new_t = (new_h - mark.size[1]) / 2
        new_mark.paste(mark, (new_l, new_t))

        mark = new_mark.rotate(rotation)

    position = determine_position(position, img, mark)

    if img.mode != 'RGBA':
        img = img.convert('RGBA')

    # make sure we have a tuple for a position now
    assert isinstance(position, tuple), 'Invalid position "%s"!' % position

    # create a transparent layer the size of the image and draw the
    # watermark in that layer.
    layer = Image.new('RGBA', img.size, (0,0,0,0))
    if tile:
        first_y = position[1] % mark.size[1] - mark.size[1]
        first_x = position[0] % mark.size[0] - mark.size[0]

        for y in range(first_y, img.size[1], mark.size[1]):
            for x in range(first_x, img.size[0], mark.size[0]):
                layer.paste(mark, (x, y))
    else:
        layer.paste(mark, position)

    # composite the watermark with the layer
    return Image.composite(layer, img, layer)


def watermark_processor(image, WATERMARK=False, **kwargs):
    if WATERMARK:
        if settings.DEBUG:
            path = finders.find(WATERMARK['image'])
        else:
            path = os.path.join(settings.STATIC_ROOT, WATERMARK['image'])
        mk = Image.open(path)

        image = watermark_filter(
                    image, mk, 
                    WATERMARK.get('position', 'BR'),
                    WATERMARK.get('opacity', 0.5),
                    WATERMARK.get('scale', 1),
                    WATERMARK.get('tile', False),
                    WATERMARK.get('greyscale', False),
                    WATERMARK.get('rotation', 0)
        )

        if WATERMARK.get('noalpha', False):
            image = image.convert('RGB')
    return image
